package beagle

import fol._
import datastructures._
import util._

package object bgtheory {

  /** Thrown when there is a conversion error from Formula to e.g. PolyAtom/Polynomial*/
  case class ConversionError(s: String) extends Message

  /**
   * All the solvers we have.
   */
  private val solvers: List[Solver] = List(
    PreprocessingSolver,
    // to be added:
    LIA.iQE.Solver,
    LIA.cooper.SolverClauses,
    LIA.cooper.SolverUnits,
    // LIA.BNBSolverUnits,
    // LIA.BNBSolverClauses,
    LIA.SMT.Z3Solver,
    LIA.SMT.CVC4Solver,
    LRA.LRASolver,
    LFA.LFASolver, 
    EmptySolver)

  /**
   * The solver we are using, set in main.scala depending on the -format flag
   */
  var solver: Solver = null // EmptySolver

  def setSolver(n: String) {
    (solvers find { _.name == n }) match {
      case Some(s) => {
	solver = s
	s.init()
      }
      case None => throw CmdlineError("no such solver: " + n)
    }
  }

  /** Thrown by some solvers if they encounter a nonlinear term */
  case class NonLinearTermFail(t: Term) extends Exception

  /** Thrown by some solvers if they encounter a term of a sort they cannot handle */
  case class IllSortedTermFail(t: Term) extends Exception

  def toSolverResult(isConsistent: Boolean) =
    if (isConsistent) SAT else UNSAT(None)
}

// Some simple operator generator classes
// All operators have to be FG. They are only dealt with by simplification, over fully ground instantiated terms
class ToIntOp(sort: Sort) extends Operator(FG, "$to_int", Arity1(sort -> bgtheory.LIA.IntSort))
class ToRatOp(sort: Sort) extends Operator(FG, "$to_rat", Arity1(sort -> bgtheory.LRA.RatSort))
class ToRealOp(sort: Sort) extends Operator(FG, "$to_real", Arity1(sort -> bgtheory.LFA.RealSort))

class IsIntOp(sort: Sort) extends Operator(FG, "$is_int", Arity1(sort -> Signature.OSort))
class IsRatOp(sort: Sort) extends Operator(FG, "$is_rat", Arity1(sort -> Signature.OSort))
class IsRealOp(sort: Sort) extends Operator(FG, "$is_real", Arity1(sort -> Signature.OSort))

class QuotientEOp(sort: Sort) extends Operator(FG, "$quotient_e", Arity2((sort, sort) -> bgtheory.LIA.IntSort))
class QuotientTOp(sort: Sort) extends Operator(FG, "$quotient_t", Arity2((sort, sort) -> bgtheory.LIA.IntSort))
class QuotientFOp(sort: Sort) extends Operator(FG, "$quotient_f", Arity2((sort, sort) -> bgtheory.LIA.IntSort))

class RemainderEOp(sort: Sort) extends Operator(FG, "$remainder_e", Arity2((sort, sort) -> bgtheory.LIA.IntSort))
class RemainderTOp(sort: Sort) extends Operator(FG, "$remainder_t", Arity2((sort, sort) -> bgtheory.LIA.IntSort))
class RemainderFOp(sort: Sort) extends Operator(FG, "$remainder_f", Arity2((sort, sort) -> bgtheory.LIA.IntSort))


/**
 * This is a FG operator used to replace $product in case a non-linear term is found.
 * It will be eliminated with simplification rules if the term ever becomes linear again.
 * TODO- perhaps give this a nicer print function?
 */
class NLPPOp(sort: Sort) extends Operator(FG, "#nlpp", Arity2((sort,sort) -> sort)) 
