package beagle.calculus

import beagle._
import datastructures._
import util._
import derivationrules._
import simplification._
import fol._
import bgtheory._

abstract class Inference {
  val name: String // The name of the used inference rule
  val bgTheory: String = "empty" // The name of the involved BG theory
}

abstract class OnePremiseInference extends Inference {
  val premise: Clause
}

abstract class TwoPremisesInference extends Inference {
  val leftPremise: Clause
  val rightPremise: Clause
}

abstract class MultiPremisesInference extends Inference {
  val premises: List[Clause]
}

case object ByInput extends Inference { val name = "input" }
case class ByRef(premise: Clause) extends OnePremiseInference { val name = "reflexivity"; override val bgTheory = "equality" }
case class ByFact(premise: Clause) extends OnePremiseInference { val name = "factorization"; override val bgTheory = "equality" }
case class BySup(leftPremise: Clause, rightPremise: Clause) extends TwoPremisesInference { val name = "superposition"; override val bgTheory = "equality" }
case class ByRes(leftPremise: Clause, rightPremise: Clause) extends TwoPremisesInference { val name = "resolution" }
case class ByChaining(leftPremise: Clause, rightPremise: Clause) extends TwoPremisesInference { val name = "chaining"; override val bgTheory = "equality" }
case class ByDefine(premise: Clause) extends OnePremiseInference { val name = "define"; override val bgTheory = "equality" }
case class BySplitLeft(premise: Clause) extends OnePremiseInference { val name = "splitLeft" }
case class BySplitRight(leftEmptyClause: Clause, premise: Clause) extends OnePremiseInference { val name = "splitRight" }
case class ByClose(premises: List[Clause], override val bgTheory: String) extends MultiPremisesInference { val name = "close" }
case class ByDemod(premises: List[Clause]) extends MultiPremisesInference { val name = "demodulation"; override val bgTheory = "equality" }
case class ByNegUnitSimp(premises: List[Clause]) extends MultiPremisesInference { val name = "negUnitSimplification" }
case class ByBGSimp(premise: Clause, override val bgTheory: String) extends OnePremiseInference { val name = "backgroundSimplification" }
case class BySimple(premise: Clause) extends OnePremiseInference { val name = "simple" }
case class ByQE(premise: Clause, override val bgTheory: String) extends OnePremiseInference { val name = "quantifierElimination" }
case class ByCNF(premise: Formula) extends Inference { val name = "cnfTransformation" }

trait ClauseOrFormula {
  val id: Int
}

object Proof {

  // Generate a proof, i.e. a sequence of inferences from input formulas and derived clauses
  // The given clause cl contains in its inference part a tree-shaped refutation
  def mkRefutation(cl: Clause): List[ClauseOrFormula] = {

    import scala.collection.mutable.ListBuffer
    var linRef = ListBuffer.empty[Clause]

    // Linearize the refutation behind the given clause
    def linearRefutation(cl: Clause) {
      // println(s"%d linearRefutation($cl)".format(cl.id))
      if (!(linRef exists (_.id == cl.id))) {
        // Loop check
        cl.inference match {
          case ByInput => linRef += cl
          case inf: ByCNF => linRef += cl
          case inf: BySplitRight => {
            linearRefutation(inf.leftEmptyClause)
            linRef += cl
          }
          case inf: OnePremiseInference => {
            linearRefutation(inf.premise)
            linRef += cl
          }
          case inf: TwoPremisesInference => {
            linearRefutation(inf.leftPremise)
            linearRefutation(inf.rightPremise)
            linRef += cl
          }
          case inf: MultiPremisesInference => {
            (inf.premises foreach { linearRefutation(_) })
            linRef += cl
          }
        }
      }
    }

    linearRefutation(cl)

    // Put "ByCNF first. Todo: Same for ByInput once used
    val (byCnfs, others) = linRef.toList partition { _.inference.isInstanceOf[ByCNF] }

    // Extract input formulas and put them before the ByCnfs
    val fs = byCnfs map { byCnf =>
      val cl = byCnf.asInstanceOf[Clause]
      cl.inference.asInstanceOf[ByCNF].premise
    }

    val all = fs ::: byCnfs ::: others

    // Remove duplicates
    var noDupls = List.empty[ClauseOrFormula]
    for (h <- all) {
      if (!(noDupls exists { f => f.isInstanceOf[Formula] == h.isInstanceOf[Formula] && f.id == h.id }))
        noDupls ::= h
    }
    noDupls.reverse // preserve order
  }

  def showRefutationTFF(ref: List[ClauseOrFormula]) {

    ref foreach {
      _ match {
        case f: Formula =>
          println("tff(%s, %s, %s, file('%s', %s)).".format(
            "f_" + f.id,
            f.role.getOrElse("axiom"),
            f, f.source.getOrElse("unknown"), f.name.getOrElse("unknown")
          ))
        case cl: Clause => {
          val bgTheoryMentioned =
            cl.inference.bgTheory match {
              case "empty" => ""
              case s => ", theory('%s')".format(s)
            }
          cl.inference match {
            case ByInput =>
              println("tff(%s, axiom, %s).".format("c_" + cl.id, cl))
            case inf: ByCNF =>
              println("tff(%s, plain, %s, inference(%s, [status(thm)], [%s])).".format("c_" + cl.id, cl, inf.name, "f_" + inf.premise.id))
            case inf: OnePremiseInference =>
              println("tff(%s, plain, %s, inference(%s, [status(thm)%s], [%s])).".format("c_" + cl.id, cl, inf.name, bgTheoryMentioned, "c_" + inf.premise.id))
            case inf: TwoPremisesInference =>
              println("tff(%s, plain, %s, inference(%s, [status(thm)%s], [%s, %s])).".format("c_" + cl.id, cl, inf.name, bgTheoryMentioned, "c_" + inf.leftPremise.id, "c_" + inf.rightPremise.id))
            case inf: MultiPremisesInference =>
              println("tff(%s, plain, %s, inference(%s, [status(thm)%s], %s)).".format("c_" + cl.id, cl, inf.name, bgTheoryMentioned,
                (inf.premises map { "c_" + _.id }).toMyString("[", ", ", "]")))
          }
        }
      }
    }
  }

  def showRefutationDefault(ref: List[ClauseOrFormula]) {
    ref foreach {
      _ match {
        case f: Formula => println("%5s: %s".format("f_" + f.id, f))
        case cl: Clause => {
          val bgTheoryMentioned =
            cl.inference.bgTheory match {
              case "empty" | "equality" => ""
              case s => s + "-"
            }
          cl.inference match {
            case ByInput => println("%5d: %s".format(cl.id, cl.toString, ByInput.name))
            case inf: ByCNF =>
              println("%5d: %s\n       by %s(%s)".format(cl.id, cl.toString, inf.name, "f_" + inf.premise.id))
            case inf: OnePremiseInference =>
              println("%5d: %s\n       by %s%s(%d)".format(cl.id, cl.toString, bgTheoryMentioned, inf.name, inf.premise.id))
            case inf: TwoPremisesInference =>
              println("%5d: %s\n       by %s%s(%d, %d)".format(cl.id, cl.toString, bgTheoryMentioned, inf.name, inf.leftPremise.id, inf.rightPremise.id))
            case inf: MultiPremisesInference =>
              println("%5d: %s\n       by %s%s%s".format(cl.id, cl.toString, bgTheoryMentioned, inf.name, (inf.premises map { _.id }).toMyString("", "(", ", ", ")")))
          }
        }
      }
    }
  }
}

