package beagle.cnfconversion

import beagle._
import fol._
import rules._

object standard {

  // Convert a formula to CNF, i.e. a list of list of Literals
  // If full is true also pure BG formulas are subject to CNF conversion,
  // which may introduce Skolem functions,
  // hence it may often be better to have full false, so that these formulas are treated as atoms by the 
  // CNF conversion and treated by QE later
  // 
  def toCNF(f: Formula, full: Boolean): List[List[Formula]] = {
    
     // println("***" + f + " --- " + full)
    
    // clumsy:
    val res = if (full) {
      
      val h = f.
	reduceInnermost(elimIffNot).
	reduceInnermost(elimIff).
	reduceInnermost(elimImpl).
	reduceInnermost(pushdownNeg).
	reduceInnermost(elimNegNeg)

        // h is a conjunction of formulas, even if a singleton.
        // Convert each element separately into CNF, which can be much more efficient.
        // In particular we save unneccessary pulling out quanifiers. A good example is SWV998=1
        // println("*** h = " + h)
        
        // Do skolemisation of outer existential if possible
        // useful in case the formula has an existential quantifier over a conjunction of
        // universally quantified inner formulas which generate skolem functions
        // in this case the h.toList(And) optimisation is blocked and we get massive skolem functions
        // eg. PUZ001+2.p yields a 17 place skolem function
         match {
          case Exists(xs, f) => {
            var sigma = Subst.empty
            xs foreach { sigma += skolemization.mkSkoSubst(List.empty, _) }
            sigma(f)
          }
          case x => x
        }
        
        h.toListAnd flatMap { _. 
        // Only now quantifiers can be pulled out, as arrows have been eliminated
        reduceInnermost(pulloutQuants).
        skolemize.
        matrix.
        // Only universal quantifiers in front now - remove
        // The matrix now is made from Neg, And, and Or,
        // convert to CNF
        reduceInnermost(pushdownNeg).
        reduceInnermost(elimNegNeg).
        reduceInnermost(pushdownOr).
        reduceInnermost(flattenAndOr).
        reduceInnermost(elimTrivial).
        toListAnd } map { _.toListOr }
    }
    else {
      //println("** here we go- not full")
      val h = f.
        reduceInnermostFG(elimIffNot).
        reduceInnermostFG(elimIff).
        reduceInnermostFG(elimImpl).
        reduceInnermost(pushdownNeg).
        reduceInnermost(elimNegNeg)

        // h is a conjunction of formulas, even if a singleton.
        // Convert each element separately into CNF, which can be much more efficient.
        // In particular we save unneccessary pulling out quanifiers. A good example is SWV998=1
        //println("*** h = " + h)
        
        // Do skolemisation of outer existential if possible
        // Peter 10/6/2014 - do *not* Skolemize if this is a BG formula
        // Rationale: cooper will treat it
         match {
          case g @ Exists(xs, f) if !g.isBG => {
            var sigma = Subst.empty
            xs foreach { sigma += skolemization.mkSkoSubst(List.empty, _) }
            sigma(f)
          }
          case x => x
        }
      
        // println("Midpoint2 "+h)
        h.toListAnd flatMap { _. 
			       // Only now quantifiers can be pulled out, as arrows have been eliminated
			       reduceInnermostFG(pulloutQuants).
			       skolemizeUnlessIsBG. // Skolemization is done in two ways: QE on variables appearing in BG only, the others in toDNF
			       matrix.
			       // Only universal quantifiers in front now - remove
			       // The matrix now is made from Neg, And, and Or,
			       // convert to CNF
			       reduceInnermostFG(pushdownNeg).
			       reduceInnermostFG(elimNegNeg).
			       reduceInnermostFG(pushdownOr).
			       reduceInnermostFG(flattenAndOr).
			       reduceInnermostFG(elimTrivial).
			       toListAnd
			     } map { _.toListOr }
    }
    //println("Result is "+res)
    res
  }
}
