package beagle.fol

import beagle._
import util._
import bgtheory._

/**
 * Foreground/background kinds, used to classify both sorts and operators
 */
sealed abstract class Kind
case object FG extends Kind
case object BG extends Kind

sealed abstract class Sort {
  // Abstract members
  val name: String
  val kind: Kind
  val params: Option[List[Sort]]
  /** some element used for "$-closing" */
  var someElement: Const = null  
  override def toString = printer.sortToString(this) //name

  //either equal or non-empty intersection
  def sortsCanUnify(s: Sort) = (this == s)
  //defines well-sortedness
  def sameUnderlyingSort(s: Sort) = (this == s)

  def isSubsortOf(s: Sort) = (this.name == s.name)

  //general case of non-intersecting sorts
  def intersectWith(s: Sort) = if (this == s) Some(s) else None

  /** A unique symbolic constant per sort, used in main loop for generating instances */
  def someSkolemElement = SymConst(Operator(kind, "#skE_"+ name, Arity0(this)))
}

case class FGSort(name: String, params: Option[List[Sort]] = None) extends Sort {
  val kind = FG
  someElement = FGConst("$" + name, this)
}

case class BGSort(name: String, params: Option[List[Sort]] = None) extends Sort {
  val kind = BG
}

/**
 * Arity: The signature of individual function (and predicate) symbols.
 */
case class Arity(argsSorts: List[Sort], resSort: Sort) {
  val nrArgs = argsSorts.length

  override def toString = printer.arityToString(this)

  def toRat = {
    def sortToRat(s: Sort) = if (s == bgtheory.LFA.RealSort) bgtheory.LRA.RatSort else s
    Arity(argsSorts map { sortToRat(_) }, sortToRat(resSort))
  }
}

object Arity0 {
  def apply(r: Sort) = new Arity(List.empty, r)
  def unapply(r: Arity) = r match {
    case Arity(Nil, t) ⇒ Some(t)
    case _ ⇒ None
  }
}
object Arity1 {
  def apply(r: (Sort, Sort)) = Arity(List(r._1), r._2)
  def unapply(r: Arity) = r match {
    case Arity(List(t1), t) ⇒ Some(t1 -> t)
    case _ ⇒ None
  }
}
object Arity2 {
  def apply(r: ((Sort, Sort), Sort)) = Arity(List(r._1._1, r._1._2), r._2)
  def unapply(r: Arity) = r match {
    case Arity(List(t1, t2), t) ⇒ Some((t1, t2) -> t)
    case _ ⇒ None
  }
}
object Arity3 {
  def apply(r: ((Sort, Sort, Sort), Sort)) = Arity(List(r._1._1, r._1._2, r._1._2), r._2)
  def unapply(r: Arity) = r match {
    case Arity(List(t1, t2, t3), t) ⇒ Some((t1, t2, t3) -> t)
    case _ ⇒ None
  }
}

/**
 * Operators: identified by their name, their kind (FG or BG) and their arity
 */
case class Operator(kind: Kind, name: String, arity: Arity) {
  override def toString = printer.opToString(this) //name

 /**
  * Used to strip off from an identifier the sort introduced by the SMT front-end.
  * Just cuts from an identifier everything from the first occurrence of ":".
  * This applies to input generated by the smttotptp utility as well.
  */
  lazy val shortName = 
    name.split(':')(0)

  lazy val isSkolem = isInstanceOf[SkolemOp]

  /** Whether the operator is for a domain element, e.g. 0,1,2 or 1/2 etc. */
  //using isDomElemOp instead
  //val isDomainOp = false

  /**
   * Default value for formatFn is functional, can be overridden of course,
   * "functional" also works with constants.
   */
  def formatFn = Operator.functional(if (beagle.util.flags.longNames.value) name else shortName)

  def toRat = Operator(kind, name, arity.toRat)

  def sort(argSorts: List[Sort] = List[Sort]()) = arity.resSort

  def compareTo(that: Operator) = 
    // first by O-Sortedness, then arity, then name
    if ((this.sort() == Signature.OSort) && (that.sort() != Signature.OSort)) 1
    else if ((that.sort() == Signature.OSort) && (this.sort() != Signature.OSort)) -1
    else if (this.arity.nrArgs > that.arity.nrArgs) 1
    else if (that.arity.nrArgs > this.arity.nrArgs) -1
    else // same arity
      this.shortName compareTo that.shortName

  /**
   * Some operators *uniquely* determine a Term, in particular
   * BG domain elements. In this case toTerm provides this Term.
   * See e.g. bgtheory/LIA/operators.scala how this is done
   */
  def toTerm: Option[Term] = None

  def isBSFGOp = kind == FG && arity.resSort.kind == BG

  // Use case:
  lazy val isDomElemOp = 
    toTerm match {
      case None => false
      case Some(t) => t.isInstanceOf[DomElem[_]]
    }

// Used in the Define inference rule via mkDefinition; need to keep a counter with parameters,
// for fairness reasons. Used in the weight function for terms
  def skolemCtr: Option[Int] = None
}

object Operator {
  // We allow operators to be equipped with FormatFn's, which
  // are functions that convert a term with that operator to a String

  // sort FormatFn = List[Term] => String
  // The most commonly used FormatFns, see Predefined.scala for usages.
  def infix[T](op: String) = (ts: List[T]) ⇒
    // "(" + ts(0) + " " + op + " " + ts(1) + ")"
    // Allow symbols be used varyadic 
    if (ts.isEmpty) "" else if (ts.tail.isEmpty) ts.head.toString else ts.toMyString("(", op, ")")
  def prefix[T](op: String) = (ts: List[T]) ⇒ op + ts(0)
  def functional[T](fun: String) = (ts: List[T]) ⇒ {
    // See if we need quoting
    val pfun =
      if (util.printer.name == "tff" && fun(0) == '#')
        "'" + fun + "'" else fun
    pfun + ts.toMyString("", "(", ", ", ")")
  }
}

/*
object DomElemOp {
  def apply(name: String, arity: Arity) =
    (new Operator(BG,name,arity) {
      override val isDomainOp = true
    })
}
*/

/**
 * JOSH- added class to distinguish skolem operators.
 * Previously the '#' prefix would guarantee minimality in the operator precedence
 * but '#' cannot be parsed by TPTP so skolem operators may not always have this.
 * TODO- does the same bug occur with define constants? Perhaps not, as they are
 * not added to the precedence like skolems?
 */
class SkolemOp(kind: Kind, name: String, arity: Arity) extends Operator(kind,name,arity)

/**
 * Operators which are polymorphic or varyadic in a theory
 * Only used in FD package
 */
class PolyOp(kind: Kind, name: String, arity: Arity) extends Operator(kind, name, arity) {
  override def equals(that: Any) = that match {
    case that: Operator ⇒ (this canEqual that) && this.name == that.name
    case _ ⇒ false
  }

  override def canEqual(other: Any): Boolean = other.isInstanceOf[PolyOp]

  override def hashCode: Int = 
    41 * ( 41 * (41 + kind.hashCode) + name.hashCode) + arity.hashCode


  /**
   * PolyOps often have generic instances which are stored in the signature. Creating a
   * PFunTerm from an Atom will use the generic instance when it is more correct to use
   * a specific instance, eg. Atom($sum,5,2) will have sort Int even though fd(7,7) is more
   * appropriate. This provides a way to create a new instance if we have found the generic
   * one.
   */
  //def newInstance(argsSorts: List[Sort]) = this
  //TODO might still need this for varyadic operators

  def sortCheck(args: List[Term]): Boolean = 
    (args.length == arity.nrArgs) &&
    ((arity.argsSorts zip args.sorts) forall { x ⇒ x._2 sameUnderlyingSort x._1 })
}

class EqualsOp(baseSort: Sort)
  extends PolyOp(baseSort.kind, "=", Arity2((baseSort, baseSort), Signature.OSort))

/**
 * A Signature consists of a set of sorts and a set of operators over these sorts.
 * Optionally, precs is a list of precedences over (the names of) operators.
 */
class Signature(val sorts: Set[Sort] = Set.empty,
  val operators: Set[Operator] = Set.empty,
  val precs: List[(String, String)] = List.empty) {

  // todo: 
  // - precs should really be a list of pairs of *operators* instead of names
  // - close precs under transitivity so that lookup becomes cheap and check for cycles

  // Some derived notions
  val bgSorts = sorts filter { _.kind == BG }
  val fgSorts = sorts filter { _.kind == FG }
  val bgOperators = operators filter { _.kind == BG }
  val fgOperators = operators filter { _.kind == FG }

  /**
    * Return a set of templates for the given sort.
    */
  def templates(s: Sort) = {
    // The FG operators with the result sort s
    val ops = fgOperators filter { _.arity.resSort == s }
    ops map { op => FGFunTerm(op,
      // Make a list of fresh variables of the proper sorts
      op.arity.argsSorts map { as =>
        as.kind match {
          case FG => GenVar("XT", Term.variantCtr.next(), as)
          case BG => if (util.flags.genvars.value) GenVar("XT", Term.variantCtr.next(), as) else AbstVar("XT", Term.variantCtr.next(), as)
        }
      })
    }
  }

  /**
   * Reusable equality operators.
   */
  val eqOperators: Map[Sort,Operator] =
    sorts.map(s => (s,Operator(FG, "$equal", Arity2((s, s) -> Signature.ISort)))).toMap
  
  /**
   * Convert all foreground operators from Real to Rat
   */
  def toRat = new Signature(sorts, bgOperators ++ (fgOperators map { _.toRat }), precs)

  /**
   * Compute the transitive closure of a given relation
   */
  def tc(rel: List[(String, String)]) = {
    var res = List.empty[(String, String)]
    // Invariant: res is transitively closed.
    // We add the elements of rel one after the other.
    // I think this is Dijkstra's algorithm (long time ago ...)
    for ((s, t) <- rel) {
      val sPred = res collect { case p if (p._2 == s) => p._1 }
      val tSucc = res collect { case p if (p._1 == t) => p._2 }
      res :::= (s, t) ::
        (for (sp <- sPred) yield (sp, t)) :::
        (for (ts <- tSucc) yield (s, ts)) :::
        (for (sp <- sPred; ts <- tSucc) yield (sp, ts))
    }
    res
  }

  val precstc = tc(precs)
  // Sanity check, for cycles
  (precstc find { case (f, g) => f == g }) match {
    case Some((f, _)) =>
      throw new GeneralError("Cycle in declared precedences, involving e.g.: " + f)
    case _ => ()
  }


  /**
 * fgOperatorsPrecMap: a linear ordering among all foreground operators,
 * using the lexicographic ordering over (a) the user given precedences and
 * (b) precedence over non-constants over constants.
 *  Moreover, $true > $false and $false is least.
 * fgOperatorsPrecMap is in fact a map from fgOperators to a set of all those fgOperators
 * that it precedes. Hence, checking whether an operator has precedence over another one should be really fast.
 * JOSH- filtered any FG skolem operators and added just before TRUE > FALSE in the
 * precedence.
 */
  lazy val fgOperatorsPrecMap: Map[Operator, Set[Operator]] = {

    var res = Map.empty[Operator, Set[Operator]]

    // The operators declared in the precedences:
    val precsDeclOps = (precs flatMap { case (f, g) => List(f, g) }).toSet
    val (skoOpsList, fgOpsList) = fgOperators.toList.partition(_.isSkolem)
    val linearized =
      // The non-zero arity operators for which we do not have precedences:
      (fgOpsList filter {
        op => op.arity.nrArgs > 0 && !(precsDeclOps contains op.shortName)
      } sortWith { (op1, op2) => (op1 compareTo op2) > 0 
      }) :::
      // The operators for which we do have precedences
      (fgOpsList filter {
        precsDeclOps contains _.shortName
      } sortWith { (op1, op2) =>
        precstc contains (op1.shortName, op2.shortName)
      }) :::
      // The zero arity operators for which we do not have precedences:
      (fgOpsList filter {
        op => op.arity.nrArgs == 0 && !(precsDeclOps contains op.shortName) &&
        op != Signature.falseOp && op != Signature.trueOp
      } sortWith { (op1, op2) => (op1.shortName compareTo op2.shortName) > 0 
      }) :::
      skoOpsList :::
      List(Signature.trueOp, Signature.falseOp)

    for (
      fs <- linearized.tails;
      if !fs.isEmpty;
      f = fs.head
    ) res += (f -> fs.tail.toSet)
    res
  }

  /**
   *  whether op1 has higher precedence than op2
   */
  def gtrPrecSimple(op1: Operator, op2: Operator) = fgOperatorsPrecMap(op1) contains op2

  // The (foreground) sorts with finite domains
  // Each key in this map is mapped to all constants ofthis finite domain
  var domain: Map[Sort, List[Term]] = null
  lazy val fdSorts = domain.keys.toList

  // New members - finding elements in a signature by a name
  def findSort(s: String) = sorts find { _.name == s }

  // def findOperator(s: String): Option[Operator] = operators find { _.name == s }
  // Find an operator by name and sorts of its arguments
  def findOperator(s: String, argsSorts: List[Sort]): Option[Operator] = {
    // println("*** operators " + operators)
    operators find {
      op ⇒
        (op.name == s && (op match {
          case op: PolyOp ⇒ true
          // That is, for VaryOps we take the first hit, assume there is just one with the same name
          case _ ⇒ {
	    /*if(s(0)=='$' && argsSorts.length!=op.arity.nrArgs) //catch duplicate keywords
	      throw new SyntaxError("Operator "+s+" with arity "+argsSorts.length+
				    " overloads defined keyword "+op.name+" with arity "+op.arity.nrArgs)*/
	    (argsSorts.length==op.arity.nrArgs) &&
            ((op.arity.argsSorts zip argsSorts) forall { x ⇒ x._2 sameUnderlyingSort x._1 })
          }
        }))
    }
  }

  /** Use this to retrieve an operator by its name, if you are sure it is not overloaded */
  def findOperator(s: String) = operators find { _.name == s }

  /*  def findFormatFn(s: String) = findOperator(s) match {
    case None     ⇒ None
    case Some(op) ⇒ Some(op.formatFn)
  }
*/

  def findFormatFn(s: String, a: Arity) =
    // Silently ignore result sort, assume operators are unique per name and args sorts
    findOperator(s, a.argsSorts) match {
      case None ⇒ None
      case Some(op) ⇒ Some(op.formatFn)
    }

  def existsOp(s: String) = operators exists { _.name == s }
  //   def apply(s: String) = findOperator(s)

  /**
   * Add a sort.
   * Note that duplicate sorts usually can't be detected, because equality of `Sort`
   * is defined (by case class) as depending on name.
   */
  def +(sort: Sort): Signature = {
    /*if (sorts exists { t ⇒ (t.name == sort.name && t != sort) })
      throw SyntaxError("double declaration of sort " + sort)*/

    if (sorts(sort)) this
    else new Signature(sorts + sort, operators, precs)
  }

  /**
   * Add a rank.
   * Note that it does not make sense to allow operators which differ only by kind,
   * e.g. adding +:(Int,Int)->Int with FG kind should not be possible.
   * @throws SyntaxError if operator with same name and arity exists in the signature.
   */
  def +(op: Operator): Signature = {
    if (operators exists { o ⇒ (o.name == op.name && o.arity == op.arity && o != op) })
      throw SyntaxError("double declaration of operator" + op)

    if (operators(op)) this else new Signature(sorts, operators + op, precs)
  }

  def analyse() {
    domain = Map.empty
    for (
      s ← fgSorts;
      if s != Signature.OSort && s != Signature.TSort;
      sOps = fgOperators filter { _.sort() == s };
      if (sOps forall { _.arity.nrArgs <= 1 }) // could also do <= 0
    ) // turn all operators into symbolic constants and set the domain to these
    {
      var domainElems = List.empty[Term]
      for (op ← sOps)
        domainElems ::=
          (if (op.arity.nrArgs == 0)
            SymConst(op)
          else // arity is 1
            PFunTerm(op, List(GenVar("X", op.arity.argsSorts.head))))
      if (!(sOps exists { _.arity.nrArgs == 0 }))
        // have no constant - add one
        domainElems ::= s.someElement
      domain += s -> domainElems
    }
  }

  def show() {
    println(printer.signatureToString(this))
  }

  /**
   * checkedXX: creates an XX, type-checked against sig and varTypes
   */
  def checkedEquation(s: Term, t: Term) = {

    if (s.sort != Signature.OSort && (s.sort sameUnderlyingSort t.sort))
      Equation(s, t) // For simplicity, no need to discover BG Equations in the parser
    else
      throw new SyntaxError("ill-sorted (dis)equation: between " + s + " and " + t)
  }

  /*  def checkedAtom(pred: String, args: List[Term]) =
    // Assume that pred has been entered into sig already
    findOperator(pred) match {
      case Some(op: VaryOp) if (op.sortCheck(args)) ⇒ Atom(pred, args)
      case Some(Operator(_, _, Arity(argsSorts, Signature.OSort))) ⇒
        if (argsSorts.zip(args map { _.sort }).forall(x ⇒ x._2 sameUnderlyingSort x._1))
          Atom(pred, args)
        else
          throw new SyntaxError("ill-sorted atom: " + Atom(pred, args))
      case _ ⇒ throw new SyntaxError("internal error: checkAtom called on non-declared predicate symbol " + pred)
    }
*/

  def checkedAtom(pred: String, args: List[Term]) =
    // Assume that pred has been entered into sig already
    findOperator(pred, args.sortsOf) match {
      case Some(op: PolyOp) if (op.sortCheck(args)) ⇒ Atom(pred, args)
      case Some(_) ⇒ Atom(pred, args)
      case None ⇒ {
        throw new SyntaxError("ill-sorted atom: " + Atom(pred, args))
      }
    }

  /*  def checkedFunTerm(fun: String, args: List[Term]): FunTerm = {
    // Assume that fun has been entered into sig already
    val op = findOperator(fun).get
    if (args.isEmpty)
      // A constant. 
      SymConst(op)
    else if (op.arity.argsSorts.zip(args map { _.sort }).forall(x ⇒ x._2 sameUnderlyingSort x._1))
      // Type Checking OK 
      PFunTerm(op, args)
    else
      throw new SyntaxError("ill-sorted term: " + PFunTerm(op, args))
  }
*/

  def checkedFunTerm(fun: String, args: List[Term]): FunTerm = {
    // Assume that fun has been entered into sig already
    findOperator(fun, args.sortsOf) match {
      case Some(op: PolyOp) => 
        if (args.isEmpty) 
          SymConst(op) 
        else if (op.sortCheck(args))
          PFunTerm(op, args)
        else
          throw SyntaxError("ill-sorted term: " + PFunTerm(op, args))
      case Some(op: Operator) =>
        if (args.isEmpty) SymConst(op) else PFunTerm(op, args)
      case None => 
        throw new SyntaxError("Undefined function symbol: " + fun + "^"+args.length)
    }
  }

  def typeExistsChecked(sort: Sort) =
    if (this.sorts contains sort)
      sort
    else
      throw new SyntaxError("Error: type has not been declared: " + sort)

  def setPrecs(newPrecs: List[(String, String)]) = 
    new Signature(sorts, operators, newPrecs)

  
  //-----------------___EXPERIMENTAL-------------------------------------------
  import datastructures._
  
  case class TempSort(name: String) extends Sort {
    val kind = FG
    val params = None
    override def sameUnderlyingSort(s: Sort) = s match {
      case _:TempSort => true
      case _ => false
    }
  }
  
  /**
   * Apply union find to $i sorts to hopefully reduce trivial inferences.
   * Invoked with -infer flag.
   */
  def inferSorts(cs: List[datastructures.Clause]): List[Clause] = {
      var sortCounter = -1
      var tempSorts = List[Sort]()
      
      def newSort = {sortCounter+=1; val ns = TempSort("S"+sortCounter); tempSorts::=ns; ns}
      
      //rebuild terms in cs with new operators given in ops
      def rebuildTerm(t: Term, ops: Map[String,Operator]): Term = {
        t match {
          case FGConst(op) if(ops.isDefinedAt(op.name)) => 
	    FGConst(ops(op.name))
          case PFunTerm(op,args) if(ops.isDefinedAt(op.name)) => 
	    PFunTerm(ops(op.name),args.map(rebuildTerm(_,ops)))
          case x => x
        }
      }
      
      def rebuild(clauseSet: List[Clause], ops: Map[String,Operator]): List[Clause] =
        for(c <- clauseSet) yield {
	  c.modified(lits = c.lits.map(_.map(rebuildTerm(_,ops))))
        }
      
      //if you have an individual sort replace with a new sort
      val newOps = 
	      for(op <- this.operators) yield {
	        Operator(op.kind,
	        		op.name,
	        Arity(op.arity.argsSorts map {case Signature.ISort => newSort; case c => c},
	        		{if (op.arity.resSort==Signature.ISort) newSort else op.arity.resSort}))
	      }
      
      //for each variable occurrence in cs, create a new sort
      var updatedCS =
	      for(c <- cs) yield {
	        var subs = Subst()
	        c.vars.foreach(x => 
	          if (x.sort==Signature.ISort) 
	            subs+=(x->new GenVar(x.name,x.index,newSort))
	        )
	        subs(c)
	      }
      
      //replace all old operators with ones having the new temporary sorts
      //BUG- when rebuilding Eqn it tries to sort using the old precedences, which fails
      Sigma = new Signature(this.sorts++tempSorts,newOps.toSet,this.precs)
      updatedCS = rebuild(updatedCS,newOps.map(op => (op.name,op)).toMap)
      
      //build equivalence classes using a map to a representative
      //initially each sort maps to itself
      var sortEq = tempSorts.zip(tempSorts).toMap
      
      //makes it easier to use sortEq
      def sortRep(s: Sort) = sortEq.getOrElse(s,s)
      
      //sort representatives should be updated too
      def updateSortEq(t: Map[Sort,Sort]) {
        sortEq++=t
        sortEq=sortEq.mapValues(s=>t.getOrElse(s,s))
      }
      //propagate sort assignments towards the leaves of terms
      def propagate(t: Term) {
        t match {
          case f: FunTerm => {
            f.op.arity.argsSorts.zip(f.args.map(_.sort)) foreach {
              case (s1,s2) if sortRep(s1)!=sortRep(s2) => updateSortEq(Map(sortRep(s2)->sortRep(s1)))
              case _ => ()
              }
            f.args.foreach(propagate _)
          }
          case x: Var => ()
        }
      }
      
      for(c <- updatedCS;
    	  l <- c.lits;
    	  eq = l.eqn){
    	  //l.eqn- both terms same sort
          if(sortRep(eq.lhs.sort)!=sortRep(eq.rhs.sort))
            updateSortEq(Map(sortRep(eq.lhs.sort)->sortRep(eq.rhs.sort)))

          propagate(eq.lhs)
          propagate(eq.rhs)
      }
      
      //these are all of the remaining sort representatives
      val properSorts = tempSorts.filter(s => sortRep(s)==s)

      val resultOps =
	(for(op <- newOps) yield {
	  (op.name -> 
	   Operator(op.kind,
	            op.name,
		    Arity(op.arity.argsSorts.map(s => sortRep(s)),
	        	  sortRep(op.arity.resSort))))
	}).toMap
      
      //apply argument sort of functions to immediate variable subterms
      def applyVarSort(t: Term, s: Sort): Term = {
        t match {
          case Var(x,i,_) => GenVar(x,i,s)
          //case _:Const => t
          case PFunTerm(op, args) => PFunTerm(op,args.zip(op.arity.argsSorts).map(a => applyVarSort(a._1,a._2)))
          case _ => t
        }
      }
      
      //replace each sort with its representative (variables too)
      Sigma = new Signature(this.sorts++properSorts,resultOps.values.toSet,this.precs)
      updatedCS = rebuild(updatedCS,resultOps)
      
      updatedCS =
	for(c <- updatedCS) yield {
	  c.modified(lits = for(l <- c.lits;
		        	eqn=l.eqn) yield {
			      Lit(l.isPositive, 
		        	  Eqn(applyVarSort(eqn.lhs,sortRep(eqn.rhs.sort)),
		        	      applyVarSort(eqn.rhs,sortRep(eqn.lhs.sort))))
			    })
	}
      
      return updatedCS
    }
}

/**
 * Predefined sorts and signatures
 */
object Signature {

  // () => 
  // Predefined sorts
  object ISort extends FGSort("$i") // Individuals (of the Herbrand Universe)
  // cannot call the object iSort because if so Scala pattern matching assumes iSort is a *variable*
  object OSort extends FGSort("$o") // Truth values , () => FGConst("$$_o", OSort)
  object TSort extends FGSort("$tType") // The sort of sorts (kinds), not really a FG sort but no problems treating it as one

  // object AnyBGSort extends BGSort("$__anyBGSort__", null)
  // object AnyFGSort extends FGSort("$__anyFGSort__") // used as the "supersort" of all sorts, used for equations

  // Predefined operators
  val trueOp = Operator(FG, "$true", Arity0(OSort))
  val falseOp = Operator(FG, "$false", Arity0(OSort))

  // Signature, over empty background signature
  var signatureEmpty = new Signature
  signatureEmpty += ISort
  signatureEmpty += OSort
  signatureEmpty += TSort // does not belong to the signature - it's meta
  // signatureEmpty += AnyFGSort

  // Operators - boolean constants
  signatureEmpty += trueOp
  signatureEmpty += falseOp

}

